# The following import is to let tests run in Python 2.7 which does not have
# unittest.mocks which was introduced in Python 3.3. Some of our projects still run
# Python 2.7 so support is necessary for now
try:
    from unittest import mock
except ImportError:
    import mock
from unittest import TestCase

from mercuryclient.api import MercuryApi


class ExperianMixinTest(TestCase):
    def setUp(self):
        self.post_api_mock = mock.patch(
            "mercuryclient.api.MercuryApi._post_json_http_request"
        ).start()
        self.addCleanup(self.post_api_mock.stop)
        self.get_api_mock = mock.patch(
            "mercuryclient.api.MercuryApi._get_json_http_request"
        ).start()
        self.addCleanup(self.get_api_mock.stop)
        self.sleep_mock = mock.patch("mercuryclient.mixins.experian.time.sleep").start()
        self.sleep_mock.return_value = None
        self.addCleanup(self.get_api_mock.stop)

    def test_request_experian_report(self):
        client = MercuryApi(
            {
                "username": "username",
                "password": "password",
                "url": "https://mercury-dev.esthenos.in",
            }
        )
        mock_response = mock.MagicMock()
        mock_response.status_code = 201
        self.post_api_mock.return_value = ("random_string", mock_response)
        application_data = {"enquiry_reason": "AUTO_LOAN", "amount_financed": 23000}
        client.request_experian_report(application_data, "some_profile")

        self.post_api_mock.assert_called_with(
            "https://mercury-dev.esthenos.in/api/v1/experian/",
            data={
                "profile": "some_profile",
                "enquiry_reason": "AUTO_LOAN",
                "amount_financed": 23000,
            },
            send_request_id=True,
            add_bearer_token=True,
        )

    def test_request_exception_raised_if_status_code_error(self):
        client = MercuryApi(
            {
                "username": "username",
                "password": "password",
                "url": "https://mercury-dev.esthenos.in",
            }
        )
        mock_response = mock.MagicMock()
        mock_response.status_code = 401
        self.post_api_mock.return_value = ("random_string", mock_response)
        application_data = {"enquiry_reason": "AUTO_LOAN", "amount_financed": 23000}
        with self.assertRaises(Exception):
            client.request_experian_report(application_data, "some_profile")

    def test_request_api_succeeds_if_status_code_success(self):
        client = MercuryApi(
            {
                "username": "username",
                "password": "password",
                "url": "https://mercury-dev.esthenos.in",
            }
        )
        mock_response = mock.MagicMock()
        mock_response.status_code = 201
        self.post_api_mock.return_value = ("random_string", mock_response)

        application_data = {"enquiry_reason": "AUTO_LOAN", "amount_financed": 23000}
        response = client.request_experian_report(application_data, "some_profile")
        self.assertEqual(response["request_id"], "random_string")
        self.assertEqual(response["status"], "Success")

    def test_response_experian_report(self):
        client = MercuryApi(
            {
                "username": "username",
                "password": "password",
                "url": "https://mercury-dev.esthenos.in",
            }
        )
        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        self.get_api_mock.return_value = ("random_string", mock_response)
        client.get_experian_response("random_string")

        self.get_api_mock.assert_called_with(
            "https://mercury-dev.esthenos.in/api/v1/experian/",
            headers={"X-Mercury-Request-Id": "random_string"},
            send_request_id=False,
            add_bearer_token=True,
        )

    def test_response_exception_raised_if_status_code_error(self):
        client = MercuryApi(
            {
                "username": "username",
                "password": "password",
                "url": "https://mercury-dev.esthenos.in",
            }
        )
        mock_response = mock.MagicMock()
        mock_response.status_code = 401
        self.get_api_mock.return_value = ("random_string", mock_response)
        with self.assertRaises(Exception):
            client.get_experian_response("random_string")

    def test_response_api_succeeds_if_status_code_success(self):
        client = MercuryApi(
            {
                "username": "username",
                "password": "password",
                "url": "https://mercury-dev.esthenos.in",
            }
        )
        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        mock_response.json = mock.MagicMock()
        mock_response.json.return_value = {"status": "SUCCESS"}
        self.get_api_mock.return_value = ("random_string", mock_response)

        response = client.get_experian_response("random_string")
        self.assertEqual(response["request_id"], "random_string")
        self.assertEqual(response["status"], "Success")
        self.assertEqual(response["response"]["status"], "SUCCESS")

    def test_entire_request_response_flow(self):
        client = MercuryApi(
            {
                "username": "username",
                "password": "password",
                "url": "https://mercury-dev.esthenos.in",
            }
        )
        mock_post_response = mock.MagicMock()
        mock_post_response.status_code = 201
        mock_get_response = mock.MagicMock()
        mock_get_response.status_code = 200
        mock_get_response.json = mock.MagicMock()
        mock_get_response.json.return_value = {"status": "SUCCESS"}
        self.post_api_mock.return_value = ("random_string", mock_post_response)
        self.get_api_mock.return_value = ("random_string", mock_get_response)

        application_data = {"enquiry_reason": "AUTO_LOAN", "amount_financed": 23000}
        response = client.fetch_experian_report(application_data, "some_profile")

        self.post_api_mock.assert_called_with(
            "https://mercury-dev.esthenos.in/api/v1/experian/",
            data={
                "profile": "some_profile",
                "enquiry_reason": "AUTO_LOAN",
                "amount_financed": 23000,
            },
            send_request_id=True,
            add_bearer_token=True,
        )
        self.get_api_mock.assert_called_with(
            "https://mercury-dev.esthenos.in/api/v1/experian/",
            headers={"X-Mercury-Request-Id": "random_string"},
            send_request_id=False,
            add_bearer_token=True,
        )
        self.sleep_mock.assert_called_with(15)
        self.assertEqual(response["status"], "SUCCESS")

    def test_entire_request_response_flow_failure(self):
        client = MercuryApi(
            {
                "username": "username",
                "password": "password",
                "url": "https://mercury-dev.esthenos.in",
            }
        )
        mock_post_response = mock.MagicMock()
        mock_post_response.status_code = 201
        mock_get_response = mock.MagicMock()
        mock_get_response.status_code = 200
        mock_get_response.json = mock.MagicMock()
        mock_get_response.json.return_value = {"status": "IN_PROGRESS"}
        self.post_api_mock.return_value = ("random_string", mock_post_response)
        self.get_api_mock.return_value = ("random_string", mock_get_response)

        application_data = {"enquiry_reason": "AUTO_LOAN", "amount_financed": 23000}
        with self.assertRaises(Exception):
            response = client.fetch_experian_report(application_data, "some_profile")



class SMSMixin:
    """
    Mixin for sending emails
    """

    def send_sms(self, recipient, message, provider, profile):
        """
        Send sms using the mercury service

        :param recipient: Phone number of recipients
        :param message: SMS message to be sent
        :param provider: SMS Provider. Currently 'kaleyra'
        :param profile: An existing profile in Mercury. The profile has to correspond
        to the provider.
        :return: (request_id, status)
        """
        api = "api/v1/sms/"
        url = "{}/{}".format(self.host, api)

        data = {
            "provider": provider,
            "profile": profile,
            "recipient": recipient,
            "message": message,
        }

        request_id, r = self._post_json_http_request(
            url, data=data, send_request_id=True, add_bearer_token=True
        )

        if r.status_code == 201:
            return {"request_id": request_id, "status": "Success"}

        raise Exception(
            "Error while sending SMS. Status: {}, Response is {}".format(
                r.status_code, r.json()
            )
        )



import time


class ExperianMixin:
    """
    Mixin for registering Experian requests
    """

    def request_experian_report(self, application_data, profile):
        """
        POST a request to Mercury to Request Experian report. This posts the request to
        Mercury and returns immediately. Use the returned request ID and poll
        get_experian_response to check for report.
        You can also use fetch_experian_report which is a helper function that combines
        this api and the result api to get the result.

        :param application_data: Dict containing application data to get Experian report
        :type application_data: dict
        :param profile: Experian profile name
        :type profile: str
        :return: Dict containing request ID and status
        :rtype: dict
        """
        api = "api/v1/experian/"
        url = "{}/{}".format(self.host, api)

        application_data["profile"] = profile

        request_id, r = self._post_json_http_request(
            url, data=application_data, send_request_id=True, add_bearer_token=True
        )

        if r.status_code == 201:
            return {"request_id": request_id, "status": "Success"}

        raise Exception(
            "Error while sending Experian request. Status: {}, Response is {}".format(
                r.status_code, r.json()
            )
        )

    def get_experian_response(self, request_id):
        """
        Get result for Experian job request at the provided request ID. You can poll
        this method to check for response. If the job is still progressing, the status
        within the response will be IN_PROGRESS and the method can continue to be polled
        until you get either a SUCCESS or FAILURE

        :param request_id: Request ID of the Experian job request
        :type request_id: str
        :return: Dict containing request ID, status and job response within "response".
            This response will contain the "status" of the job (IN_PROGRESS, SUCCESS or
            FAILURE), a "message" and a "data" key containing a dict with the bureau
            report
        :rtype: dict
        """
        api = "api/v1/experian/"
        url = "{}/{}".format(self.host, api)

        request_id, r = self._get_json_http_request(
            url,
            headers={"X-Mercury-Request-Id": request_id},
            send_request_id=False,
            add_bearer_token=True,
        )

        if r.status_code == 200:
            return {"request_id": request_id, "status": "Success", "response": r.json()}

        raise Exception(
            "Error while getting Experian response. Status: {}, Response is {}".format(
                r.status_code, r.json()
            )
        )

    def fetch_experian_report(
        self, application_data, profile, max_attempts=8, retry_backoff=15
    ):
        """
        fetch_experian_report Generate an Experian request and get job result

        :param application_data: Dict containing application data for Experian request
        :type application_data: dict
        :param profile: Experian profile name
        :type profile: str
        :param max_attempts: Number of attempts to make when fetching the result,
            defaults to 8
        :type max_attempts: int, optional
        :param retry_backoff: Number of seconds to backoff when retrying to get the
            result, defaults to 15
        :type retry_backoff: int, optional
        :return: Dict containing the job result
        :rtype: dict
        """

        response = self.request_experian_report(application_data, profile)

        request_id = response["request_id"]

        attempts = 0
        while attempts < max_attempts:
            time.sleep(retry_backoff)
            result = self.get_experian_response(request_id)
            if result["response"].get("status") != "IN_PROGRESS":
                return result["response"]

            retry_backoff *= 2
            attempts += 1

        raise Exception("Error while getting Experian response. Status: IN_PROGRESS")
